#include <qpixmap.h>
#include <qstyle.h>
#include <qpainter.h>
#include <qpushbutton.h>
#include <qimage.h>
#include <qstylefactory.h>
#include <kapplication.h>
#include <kcmdlineargs.h>
#include <qtabbar.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qscrollbar.h>
#include <qmenubar.h>
#include <qmenudata.h>
#include <qcombobox.h>
#include <qprogressbar.h>
#include <qslider.h>
#include <qspinbox.h> 
#include <qtoolbutton.h>
#include <kglobalsettings.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <gdk/gdkx.h>

#include <queue>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>

#include <gtk/gtknotebook.h>
#include <gtk/gtkgc.h>
#include <gtk/gtkstyle.h>

#include "qt_qt_wrapper.h"
#include "qt_style.h"


bool gtkQtEnable = false;

QScrollBar* meepBar    = 0;    // Scrollbars are strange beasts - we only use one instance
QWidget* meepWidget = 0;
QSlider* meepSlider = 0;
QTabBar* meepTabBar = 0;
GdkGC* altBackGC = 0;

/* See below */
typedef std::deque<CachedImage> cache_vec;
cache_vec cache;

const QPixmap* backgroundTile;

CachedImage::CachedImage(WidgetType t, GtkStateType s, int w, int h, QImage i, unsigned flags_, float percentage_)
 : type(t), state(s), w(w), h(h), image(i), flags(flags_), percentage(percentage_) {
}

/* The original cacheAppend used a QValueList (eg, linked list).
 * std::deque seems purpose-built for this.  It has fast insertion and deletion on both
 * ends,  and cache finds have better locality of reference than a linked list.  As far
 * as I can tell, the QTL has nothing that compares, and a decent implementation is
 * nontrivial.  And we're paying for it anyway on on most compilers, may as well use it.
 * */
void cacheAppend(CachedImage image) {
	cache_vec::size_type size = cache.size();
	while (size > 50) {
		cache.pop_front();
		--size;
	}
	cache.push_back(image);
}

/* Finds an item in cache matching all attribs.  If you want to test specific flags, pass them. 
 * The original linear search that was scattered throughout the code searched oldest->newest.
 * This version implements something closer to a most-recently-used queue.  This is especially
 * important as the cache size is pretty large; so older entries probably arent even around
 * any more; or, best case, not on the current dialog, which is most likely to get repaint
 * events. 
 * */
cache_vec::value_type* cacheFind(CachedImage::WidgetType type, GtkStateType state, int w, int h, int flags = 0, float percentage = 0) {
	for(cache_vec::reverse_iterator it = cache.rbegin(); it != cache.rend(); ++it)
		if(it->type == type && it->state == state
			&& it->w == w && it->h == h
			  && it->flags == flags && it->percentage == percentage) 
				return &(*it);
	return 0;
}

		
/* Now to get rid of a ton of un-needed new's across the board.  `new' and `delete' are 
 * non-trivial operations.  You normally just don't notice it; until you're painting a window
 * with 50 widgets, with each paint operation requiring 3-4 news and 3-4 delete's.  The cost
 * of indirection is `not insubstantial'. */

 
static int dummy_x_errhandler( Display *dpy, XErrorEvent *err )
{
	return 0;
}
static int dummy_xio_errhandler( Display * )
{
	return 0;
}


void createQApp() {
	int argc = 1;
	char** argv;
	// Supply it with fake data to keep KApplication happy
	argv = (char**) malloc(sizeof(char*));
	argv[0] = (char*) malloc(sizeof(char) * 19);
	strncpy(argv[0], "gtk-qt-application", 19);
	
	QCString cmdlinePath;
	QCString cmdLine;
	cmdlinePath.sprintf("/proc/%d/cmdline", getpid());
	int fd = open(cmdlinePath, O_RDONLY);
	if (fd == -1)
	{
		printf("Gtk-Qt theme engine warning:\n");
		printf("  Could not open %s\n", (const char*)cmdlinePath);
		printf("  This may cause problems for the GNOME window manager\n");
	}
	else
	{
		while (1)
		{
			char data[80];
			int len = read(fd, data, 80);
			if (len == 0)
				break;
			cmdLine += data;
		}
		close(fd);
	}
	
	char* sessionEnv = getenv("SESSION_MANAGER");
	if (QString(sessionEnv).endsWith(QString::number(getpid())) || cmdLine.contains("gnome-wm") || cmdLine.contains("metacity"))
	{
		printf("Not initalising the Gtk-Qt theme engine\n");
	}
	else
	{
		int (*original_x_errhandler)( Display *dpy, XErrorEvent * );
		int (*original_xio_errhandler)( Display *dpy );
		original_x_errhandler = XSetErrorHandler( dummy_x_errhandler );
		original_xio_errhandler = XSetIOErrorHandler( dummy_xio_errhandler );
	
		unsetenv("SESSION_MANAGER");
		//new QApplication(gdk_x11_get_default_xdisplay());
		KCmdLineArgs::init(argc, argv, "gtk-qt-application", "GTK Application", "", "", true);
		new KApplication(argc, argv);
		setenv("SESSION_MANAGER", sessionEnv, 1);
		
		XSetErrorHandler( original_x_errhandler );
		XSetIOErrorHandler( original_xio_errhandler );
		
		gtkQtEnable = true;
	}
	
	if (!gtkQtEnable)
		return;
	
	meepBar    = new QScrollBar(0);
	meepWidget = new QWidget(0);
	meepSlider = new QSlider(meepWidget);
	meepWidget->polish();
	
	meepTabBar = new QTabBar(meepWidget);
	QTab* leftDummyTab = new QTab;
	QTab* rightDummyTab = new QTab;
	meepTabBar->addTab(leftDummyTab);
	meepTabBar->addTab(rightDummyTab);

	backgroundTile = meepWidget->paletteBackgroundPixmap();
}

void destroyQApp()
{
	if (!gtkQtEnable)
		return;
	delete meepBar;
	delete meepWidget;
	delete qApp;
	if (altBackGC != 0)
		gtk_gc_release(altBackGC);
}

QStyle::SFlags stateToSFlags(GtkStateType state)
{
	switch (state)
	{
		case GTK_STATE_ACTIVE:
			return QStyle::Style_Enabled | QStyle::Style_Down;
		case GTK_STATE_PRELIGHT:
			return QStyle::Style_Enabled | QStyle::Style_MouseOver;
		case GTK_STATE_SELECTED:
			return QStyle::Style_Enabled | QStyle::Style_HasFocus;
		case GTK_STATE_INSENSITIVE:
			return QStyle::Style_Default;
		default:
			return QStyle::Style_Enabled;
	}
}


// Converts a QImage into an RGB format understood by GTK
// Works on BigEndian machines (PPC)
// Jonas Borgstr� <jonas@edgewall.com>

void rearrangeColorComponents(QImage &image)
{
	if (!gtkQtEnable)
		return;
	
	if (QImage::systemBitOrder() == QImage::LittleEndian)
	{
		image = image.swapRGB();
		return;
	}
	
	
	int height = image.height();

	for(int line = 0; line < height; line++) {
		guint32 *pixel = (guint32 *)image.scanLine(line);
		QRgb *rgb = (QRgb *)pixel;
		int width = image.width();
		while (width--) {
			*(pixel) = (qRed(*rgb) & 0xff)  << 24 |
				(qGreen(*rgb) & 0xff) << 16 |
				(qBlue(*rgb) & 0xff)  <<  8 |
				(qAlpha(*rgb) & 0xff);
			pixel++;
			rgb++;
		}
	}
}


// The drawing functions follow the same pattern:
//  * Check for a cached pixmap of this widget (if one exists, use it)
//  * Set the appropriate flags
//  * Ask QT to paint the widget to a pixmap
//  * Convert the pixmap to a BGR image
//  * Add this image to the cache
//  * Ask GDK to paint the image to the window


void drawButton(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::Button, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap     pixmap(w, h);
	QPainter    painter(&pixmap);
	QPushButton button(0);

	QStyle::SFlags sflags = stateToSFlags(state);

	painter.fillRect(0,0,w,h,qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_PushButton, &painter, &button, 
	                          QRect(0,0,w,h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::Button, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawTab(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::Tab, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);
	
	// GTK doesn't tell us if our tab is on the left, right, or middle of the tabbar
	// So, let's always assume it's in the middle - it looks pretty
	QTab* tab = new QTab;
	meepTabBar->insertTab(tab,1);

	QStyle::SFlags sflags = stateToSFlags(state);

	if (state != GTK_STATE_ACTIVE)
		sflags = QStyle::Style_Selected;

	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_TabBarTab, &painter, meepTabBar, QRect(0,0,w,h), qApp->palette().active(), sflags, QStyleOption(tab));
	
	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);
	
	cacheAppend(CachedImage(CachedImage::Tab, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
	meepTabBar->removeTab(tab);
}

void drawVLine(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int ySource, int yDest)
{
	if (!gtkQtEnable)
		return;
	
	int width = style->xthickness;
	int height = abs(ySource-yDest);

	QPixmap pixmap(width, height);
	QPainter painter(&pixmap);

	painter.fillRect(2, 0, width - 2, height, qApp->palette().active().background());
	painter.setPen( qApp->palette().active().mid() );
	painter.drawLine( 0, 0, 0, height );
	painter.setPen( qApp->palette().active().light() );
	painter.drawLine( 1, 0, 1, height );

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, ySource, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawHLine(GdkWindow* window, GtkStyle* style, GtkStateType state, int y, int xSource, int xDest)
{
	if (!gtkQtEnable)
		return;
	
	int width = abs(xSource-xDest);
	int height = style->ythickness;

	QPixmap pixmap(width, height);
	QPainter painter(&pixmap);

	painter.fillRect(0, 2, width, height-2, qApp->palette().active().background());
	painter.setPen(qApp->palette().active().mid() );
	painter.drawLine(0, 0, width, 0);
	painter.setPen(qApp->palette().active().light());
	painter.drawLine(0, 1, width, 1);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	gdk_draw_rgb_32_image(window, style->bg_gc[state], xSource, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawLineEdit(GdkWindow* window, GtkStyle* style, GtkStateType state, int hasFocus, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::LineEdit, state, w, h, (hasFocus ? CachedImage::has_focus : 0)))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w, h);
	QPainter painter(&pixmap);

	QStyle::SFlags sflags = stateToSFlags(state);
	if (hasFocus)
		sflags |= QStyle::Style_HasFocus;
	
	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawPrimitive(QStyle::PE_PanelLineEdit, &painter, QRect(0, 0, w, h), qApp->palette().active(), sflags, QStyleOption(1,1));

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::LineEdit, state, w, h, image, hasFocus ? CachedImage::has_focus : 0));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawFrame(GdkWindow* window, GtkStyle* style, GtkStateType state, GtkShadowType shadow_type, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::Frame, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w, h);
	QPainter painter(&pixmap);

	QStyle::SFlags sflags = stateToSFlags(state);
	if ((shadow_type == GTK_SHADOW_IN) || (shadow_type == GTK_SHADOW_ETCHED_IN))
		sflags |= QStyle::Style_Sunken;

	if ((backgroundTile) && (!backgroundTile->isNull()))
		painter.fillRect(0, 0, w, h, QBrush(QColor(255,255,255), *backgroundTile));
	else
		painter.fillRect(0, 0, w, h, qApp->palette().active().background());

	qApp->style().drawPrimitive(QStyle::PE_Panel, &painter, QRect(0, 0, w, h), qApp->palette().active(), sflags, QStyleOption(2,2) );

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::Frame, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawComboBox(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::ComboBox, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);
	QComboBox cb(false, 0);
	cb.resize(w,h);
	
	QStyle::SFlags sflags = stateToSFlags(state);
	QStyle::SCFlags scflags = QStyle::SC_ComboBoxArrow | QStyle::SC_ComboBoxFrame | QStyle::SC_ComboBoxListBoxPopup;
	QStyle::SCFlags activeflags = QStyle::SC_None;

	if (state == GTK_STATE_PRELIGHT)
		activeflags = QStyle::Style_MouseOver;

	painter.fillRect(0,0, w, h, qApp->palette().active().background());
	qApp->style().drawComplexControl(QStyle::CC_ComboBox, &painter, &cb, QRect(0, 0, w, h), qApp->palette().active(), sflags, scflags, activeflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::ComboBox, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawCheckBox(GdkWindow* window, GtkStyle* style, GtkStateType state, int checked, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::CheckBox, state, w, h, checked ? CachedImage::is_checked : 0))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);
	QCheckBox checkbox(0);

	QStyle::SFlags sflags = stateToSFlags(state);
	sflags |= (checked ? QStyle::Style_On : QStyle::Style_Off);

	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_CheckBox, &painter, &checkbox, QRect(0, 0, w, h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::CheckBox, state, w, h, image, checked ? CachedImage::is_checked : 0));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, 
		                  GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawMenuCheck(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::MenuCheck, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);

	QStyle::SFlags sflags = stateToSFlags(state);
	
	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawPrimitive(QStyle::PE_CheckMark, &painter, QRect(0, 0, w, h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::MenuCheck, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, 
		                  GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawRadioButton(GdkWindow* window, GtkStyle* style, GtkStateType state, int checked, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::RadioButton, state, w, h, checked ? CachedImage::is_checked : 0))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);
	QRadioButton radio(0);

	QStyle::SFlags sflags = stateToSFlags(state);
	sflags |= checked ? QStyle::Style_On : QStyle::Style_Off;

	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_RadioButton, &painter, &radio, QRect(0,0,w,h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::RadioButton, state, w, h, image, checked ? CachedImage::is_checked : 0));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawScrollBar(GdkWindow* window, GtkStyle* style, GtkStateType state, int orientation, GtkAdjustment* adj, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	QScrollBar sbar(NULL);
	sbar.resize(w,h);
	
	sbar.setMinValue((int)adj->lower);

	// Patch from Chad Kitching <chadk@cmanitoba.com>
	sbar.setMaxValue((int)(adj->upper - adj->page_increment - adj->step_increment));
	sbar.setValue((int)adj->value);

	sbar.setPageStep((int)adj->page_size);
	sbar.setOrientation(orientation ? Qt::Vertical : Qt::Horizontal);
	
	QStyle::SFlags sflags = stateToSFlags(state);
	if (sflags |= QStyle::Style_Down) sflags = QStyle::Style_Enabled;
	if (orientation == GTK_ORIENTATION_HORIZONTAL) sflags |= QStyle::Style_Horizontal;
	
	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);
	
	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawComplexControl(QStyle::CC_ScrollBar, &painter, &sbar, QRect(0, 0, w, h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawToolButton(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
		
	if(CachedImage* it = cacheFind(CachedImage::ToolButton, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}
	
	QToolButton button(NULL);
	
	QStyle::SFlags sflags = stateToSFlags(state);
	QStyle::SCFlags activeflags = QStyle::SC_None;
	if (state == GTK_STATE_ACTIVE)
	{
		sflags |= QStyle::Style_AutoRaise;
		activeflags = QStyle::SC_ToolButton;
	}
	else
		sflags |= QStyle::Style_AutoRaise | QStyle::Style_Raised;
	
	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);
	
	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawComplexControl(QStyle::CC_ToolButton, &painter, &button, QRect(0, 0, w, h), qApp->palette().active(), sflags, QStyle::SC_ToolButton, activeflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);
	
	cacheAppend(CachedImage(CachedImage::ToolButton, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawMenuBarItem(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::MenuBarItem, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w, h);
	QPainter painter(&pixmap);
	QMenuItem mi;
	QMenuBar mb(0);
	
	QStyle::SFlags sflags = QStyle::Style_Down | QStyle::Style_Enabled | QStyle::Style_Active | QStyle::Style_HasFocus;

	qApp->style().drawControl(QStyle::CE_MenuBarItem, &painter, &mb, QRect(0, 0, w, h), qApp->palette().active(), sflags, QStyleOption(&mi));

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::MenuBarItem, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawMenuItem(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::MenuItem, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);
	QPopupMenu pm;
	QMenuData md;
	QMenuItem* mi = md.findItem(md.insertItem(""));
	
	QStyleOption opt(mi, 16, 16);
	QStyle::SFlags sflags = QStyle::Style_Active | QStyle::Style_Enabled;

	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_PopupMenuItem, &painter, &pm, QRect(0,0,w,h), qApp->palette().active(), sflags, opt);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::MenuItem, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawSplitter(GdkWindow* window, GtkStyle* style, GtkStateType state, int orientation, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::Splitter, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);

	QStyle::SFlags sflags = stateToSFlags(state);
	if (orientation == GTK_ORIENTATION_HORIZONTAL) sflags |= QStyle::Style_Horizontal;

	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawPrimitive(QStyle::PE_Splitter, &painter, QRect(0,0,w,h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::Splitter, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawTabFrame(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::TabFrame, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w, h);
	QPainter painter(&pixmap);
	QStyleOption opt(2, 2); // line width

	QStyle::SFlags sflags = stateToSFlags(state);

	if ((backgroundTile) && (!backgroundTile->isNull()))
		painter.fillRect(0, 0, w, h, QBrush(QColor(255,255,255), *backgroundTile));
	else
		painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawPrimitive(QStyle::PE_PanelTabWidget, &painter, QRect(0,0,w,h), qApp->palette().active(), sflags, opt);
	
	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::TabFrame, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawMenu(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::Menu, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w, h);
	QPainter painter(&pixmap);
	QStyle::SFlags sflags = stateToSFlags(state);

	qApp->style().drawPrimitive(QStyle::PE_PanelPopup, &painter, QRect(0,0,w,h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::Menu, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawProgressBar(GdkWindow * window, GtkStyle * style, GtkStateType state, gfloat percentage, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::ProgressBar, state, w, h, 0, percentage))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}
	
	if ((w<=1) || (h<=1))
		return; // Trying to draw something that small caused a segdault
		

	QProgressBar bar(100, NULL);
	bar.resize(w,h);
	bar.setProgress((int)(percentage*100.0));
	bar.setCenterIndicator(false);
	bar.setPercentageVisible(false);

	QImage image = QPixmap::grabWidget(&bar).convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::ProgressBar, state, w, h, image, 0, percentage));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawSlider(GdkWindow * window, GtkStyle * style, GtkStateType state, GtkAdjustment* adj, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if (CachedImage* it = cacheFind(CachedImage::Slider, state, w, h, 0,  ((adj->value-adj->lower)/(adj->upper-adj->lower))))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w, h);
	QPainter painter(&pixmap);
	meepSlider->setMinValue(0);
	meepSlider->setMaxValue(100);
	meepSlider->setValue((int)((adj->value-adj->lower)/(adj->upper-adj->lower)*100));
	meepSlider->resize(w, h);
	meepSlider->setOrientation((w>h) ? Qt::Horizontal : Qt::Vertical);
	meepSlider->setEnabled(state != GTK_STATE_INSENSITIVE);

	QImage image = QPixmap::grabWidget(meepSlider).convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::Slider, state, w, h, image, 0, ((adj->value-adj->lower)/(adj->upper-adj->lower))));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawSpinButton(GdkWindow * window, GtkStyle * style, GtkStateType state, int direction, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
		
	if(CachedImage* it = cacheFind(CachedImage::SpinButton, state, w, h, direction))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}
	
	
	QPixmap pixmap(w+100, h*2);
	QPainter painter(&pixmap);
	QSpinWidget spinWidget(NULL);
	spinWidget.resize(w+100,h*2);

	QStyle::SFlags sflags = stateToSFlags(state);
	
	painter.fillRect(0, 0, w+100, h*2, qApp->palette().active().background());
	qApp->style().drawComplexControl(QStyle::CC_SpinWidget, &painter, &spinWidget, QRect(0, 0, w+100, h*2), qApp->palette().active(), sflags, QStyle::SC_SpinWidgetUp | QStyle::SC_SpinWidgetDown | QStyle::SC_SpinWidgetButtonField, (state == GTK_STATE_ACTIVE) ? QStyle::SC_SpinWidgetUp | QStyle::SC_SpinWidgetDown : QStyle::SC_None);
	
	QPixmap croppedPixmap(w, h);
	croppedPixmap.fill(qApp->palette().active().background());
	QRect spinSize = qApp->style().querySubControlMetrics(QStyle::CC_SpinWidget, &spinWidget, QStyle::SC_SpinWidgetButtonField );
	copyBlt(&croppedPixmap, 0, 0, &pixmap, spinSize.x(), direction*h, w, h);
	
	QImage image = croppedPixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::SpinButton, state, w, h, image, direction));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawArrow(GdkWindow* window, GtkStyle* style, GtkStateType state, GtkArrowType direction, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::Arrow, state, w, h, direction))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);

	QStyle::SFlags sflags = stateToSFlags(state);
	QStyle::PrimitiveElement element;
	switch(direction)
	{
	case GTK_ARROW_UP:    element = QStyle::PE_ArrowUp;    break;
	case GTK_ARROW_DOWN:  element = QStyle::PE_ArrowDown;  break;
	case GTK_ARROW_LEFT:  element = QStyle::PE_ArrowLeft;  break;
	case GTK_ARROW_RIGHT: element = QStyle::PE_ArrowRight; break;
	}

	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawPrimitive(element, &painter, QRect(0,0,w,h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::Arrow, state, w, h, image, direction));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawListHeader(GdkWindow* window, GtkStyle* style, GtkStateType state, int x, int y, int w, int h)
{
	if (!gtkQtEnable)
		return;
	
	if(CachedImage* it = cacheFind(CachedImage::ListHeader, state, w, h))
	{
		gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, it->image.bits(), it->image.bytesPerLine());
		return;
	}

	QPixmap pixmap(w,h);
	QPainter painter(&pixmap);

	QStyle::SFlags sflags = stateToSFlags(state) | QStyle::Style_Horizontal;

	painter.fillRect(0, 0, w, h, qApp->palette().active().background());
	qApp->style().drawPrimitive(QStyle::PE_HeaderSection, &painter, QRect(0,0,w,h), qApp->palette().active(), sflags);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);

	cacheAppend(CachedImage(CachedImage::ListHeader, state, w, h, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state], x, y, w, h, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

GdkGC* alternateBackgroundGc(GtkStyle* style)
{
	// Alternate background color for listviews
	if (altBackGC != 0)
		return altBackGC;
	
	GdkColor altBackColor;
	QColor meep = KGlobalSettings::alternateBackgroundColor();
	mapColour(&altBackColor, meep);
	gdk_colormap_alloc_color(style->colormap, &altBackColor, FALSE, TRUE);
	
	GdkGCValues gc_values;
	GdkGCValuesMask gc_values_mask;
	gc_values_mask = GDK_GC_FOREGROUND;
	gc_values.foreground = altBackColor;
	
	altBackGC = (GdkGC*) gtk_gc_get (style->depth, style->colormap, &gc_values, gc_values_mask);
	
	return altBackGC;
}


// Thanks Martin Dvorak of metatheme
void parse_rc_string(QString qdefs, QString qpattern)
{
	const char* defs = qdefs.latin1();
	const char* pattern = qpattern.latin1();
	static int dynamic_counter = 0;
	GString *str;
	
	str = g_string_new(NULL);
	g_string_printf(str, "style \"gtk-qt-dynamic-%d\" { %s } widget_class \"%s\" style \"gtk-qt-dynamic-%d\"",
			dynamic_counter, defs, pattern, dynamic_counter);
	++dynamic_counter;
	gtk_rc_parse_string(str->str);
	g_string_free(str, TRUE);
}

void setRcProperties(GtkRcStyle* rc_style)
{
	if (!gtkQtEnable)
		return;
	
	parse_rc_string("GtkScrollbar::has-secondary-backward-stepper = 1", "*");
	parse_rc_string("GtkScrollbar::min-slider-length = " + QString::number(qApp->style().pixelMetric(QStyle::PM_ScrollBarSliderMin)), "*");
	parse_rc_string("GtkScrollbar::slider-width = " + QString::number(qApp->style().pixelMetric(QStyle::PM_ScrollBarExtent)), "*");
	parse_rc_string("GtkButton::child-displacement-x = " + QString::number(qApp->style().pixelMetric(QStyle::PM_ButtonShiftHorizontal)), "*");
	parse_rc_string("GtkButton::child-displacement-y = " + QString::number(qApp->style().pixelMetric(QStyle::PM_ButtonShiftVertical)), "*");
	QSlider slider(NULL); // To keep BlueCurve happy
	parse_rc_string("GtkScale::slider-length = " + QString::number(qApp->style().pixelMetric(QStyle::PM_SliderLength, &slider)), "*");
	
	parse_rc_string("xthickness = " + QString::number(qApp->style().pixelMetric(QStyle::PM_DefaultFrameWidth)), "*.GtkMenu");
	parse_rc_string("ythickness = " + QString::number(qApp->style().pixelMetric(QStyle::PM_DefaultFrameWidth)), "*.GtkMenu");
	
	parse_rc_string("xthickness = 5", "*.GtkMenu.Gtk*MenuItem");
	
	// This one may not work...
	//insertIntProperty(rc_style, "GtkCheckButton", "indicator-size", qApp->style().pixelMetric(QStyle::PM_IndicatorHeight) );
}

void setMenuBackground(GtkStyle* style)
{
	if (!gtkQtEnable)
		return;
	
	// Get the menu background image
	QPixmap pixmap(1024, 25); // Meh
	QPainter painter(&pixmap);
	QPopupMenu pm;
	QMenuData md;
	QMenuItem* mi = md.findItem(md.insertItem(""));
	
	QStyleOption opt(mi, 16, 16);
	QStyle::SFlags sflags = QStyle::Style_Default;

	if ((backgroundTile) && (!backgroundTile->isNull()))
		painter.fillRect(0, 0, 1024, 25, QBrush(QColor(255,255,255), *backgroundTile));
	else
		painter.fillRect(0, 0, 1024, 25, qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_PopupMenuItem, &painter, &pm, QRect(0,0,1024,25), qApp->palette().active(), sflags, opt);

	QImage image = pixmap.convertToImage();
	rearrangeColorComponents(image);
	
	GdkPixmap* gpixmap = gdk_pixmap_new(0, image.width(), image.height(), style->depth);
	GdkGCValuesMask gc_values_mask = GDK_GC_FOREGROUND;
	GdkGCValues gc_values;
	gc_values.foreground = style->fg[0];
	GdkGC* gc = gtk_gc_get(style->depth, style->colormap, &gc_values, gc_values_mask);
	gdk_draw_rgb_32_image(gpixmap, gc, 0, 0, image.width(), image.height(), GDK_RGB_DITHER_NONE, 
					image.bits(), image.bytesPerLine());
	if (gpixmap) 
		QTENGINE_STYLE(style)->menuBackground = gpixmap;
	gtk_gc_release(gc);
}

// It has a 'u' damnit
void mapColour(GdkColor* g, QColor q)
{
	g->red = q.red() * 257;
	g->green = q.green() * 257;
	g->blue = q.blue() * 257;
}

void setColors(GtkStyle* style)
{
	if (!gtkQtEnable)
		return;
	
	bool useBg = ((backgroundTile) && (!backgroundTile->isNull()));
	QImage image;

	if (useBg)
	{
		image = backgroundTile->convertToImage();
		rearrangeColorComponents(image);
	}

	for (uint i=0 ; i<5 ; i++)
	{
		switch (i)
		{
		case GTK_STATE_NORMAL:
		case GTK_STATE_INSENSITIVE:
		case GTK_STATE_ACTIVE:
			mapColour(&style->bg[i], qApp->palette().active().background());
			break;
		case GTK_STATE_PRELIGHT:
		case GTK_STATE_SELECTED:
			mapColour(&style->bg[i], qApp->palette().active().highlight());
			break;
		}
		
		mapColour(&style->fg[i], qApp->palette().active().foreground());
		
		if (i == GTK_STATE_SELECTED)
		{
			mapColour(&style->base[i], qApp->palette().active().highlight());
			mapColour(&style->text[i], qApp->palette().active().highlightedText());
		}
		else if (i == GTK_STATE_INSENSITIVE)
		{
			mapColour(&style->base[i], qApp->palette().disabled().highlight());
			mapColour(&style->base[i], qApp->palette().disabled().highlightedText());
			mapColour(&style->fg[i], qApp->palette().disabled().foreground());
		}
		else
		{
			mapColour(&style->base[i], qApp->palette().active().base());
			mapColour(&style->text[i], qApp->palette().active().text());
		}
	}

	if (useBg)
	{
		GdkPixmap* pixmap = gdk_pixmap_new(0, image.width(), image.height(), style->depth);
		GdkGCValuesMask gc_values_mask = GDK_GC_FOREGROUND;
		GdkGCValues gc_values;
		gc_values.foreground = style->fg[0];
		GdkGC* gc = gtk_gc_get(style->depth, style->colormap, &gc_values, gc_values_mask);
		gdk_draw_rgb_32_image(pixmap, gc, 0, 0, image.width(), image.height(), GDK_RGB_DITHER_NONE, 
				               image.bits(), image.bytesPerLine());
		if (pixmap) 
			style->bg_pixmap[GTK_STATE_NORMAL] = pixmap;
		gtk_gc_release(gc);
	}
	
	setMenuBackground(style);
}


